import abc
from typing import *

import hail as hl
from hail.expr.expressions import Expression, ExpressionException, to_expr
from hail.expr.types import *
from hail.typecheck import TypeChecker, TypecheckFailure
from hail.utils.java import escape_parsable

__all__ = [
    'expr_any',
    'expr_int32',
    'expr_int64',
    'expr_float32',
    'expr_float64',
    'expr_call',
    'expr_bool',
    'expr_str',
    'expr_locus',
    'expr_interval',
    'expr_array',
    'expr_ndarray',
    'expr_set',
    'expr_dict',
    'expr_tuple',
    'expr_struct',
    'expr_oneof',
    'expr_numeric',
    'coercer_from_dtype',
]

T = TypeVar('T')


class ExprCoercer(TypeChecker):
    __metaclass__ = abc.ABCMeta

    @property
    @abc.abstractmethod
    def str_t(self) -> str:
        ...

    def requires_conversion(self, t: HailType) -> bool:
        assert self.can_coerce(t), t
        return self._requires_conversion(t)

    @abc.abstractmethod
    def _requires_conversion(self, t: HailType) -> bool:
        """Assumes that :meth:`can_coerce` is ``True``"""
        ...

    @abc.abstractmethod
    def can_coerce(self, t: HailType) -> bool:
        ...

    def coerce(self, x) -> Expression:
        x = to_expr(x)
        if not self.can_coerce(x.dtype):
            raise ExpressionException(f"cannot coerce type '{x.dtype}' to type '{self.str_t}'")
        if self._requires_conversion(x.dtype):
            return self._coerce(x)
        else:
            return x

    def _coerce(self, x: Expression) -> Expression:
        # many coercers don't write any coercion rules, so don't override
        raise AssertionError()

    def expects(self):
        return f"expression of type {self.str_t}"

    def check(self, x: Any, caller: str, param: str) -> Any:
        try:
            return self.coerce(to_expr(x))
        except ExpressionException as e:
            raise TypecheckFailure from e

    def format(self, arg):
        if isinstance(arg, Expression):
            return str(arg)
        else:
            return super(ExprCoercer, self).format(arg)


class AnyCoercer(ExprCoercer):
    @property
    def str_t(self):
        return 'any'

    def _requires_conversion(self, t: HailType) -> bool:
        return False

    def can_coerce(self, t: HailType) -> bool:
        return True


class BoolCoercer(ExprCoercer):
    @property
    def str_t(self):
        return 'bool'

    def _requires_conversion(self, t: HailType) -> bool:
        return False

    def can_coerce(self, t: HailType) -> bool:
        return t == tbool


class Int32Coercer(ExprCoercer):
    @property
    def str_t(self):
        return 'int32'

    def _requires_conversion(self, t: HailType) -> bool:
        return t != tint32

    def can_coerce(self, t: HailType) -> bool:
        return t in (tbool, tint32)

    def _coerce(self, x):
        return x._method("toInt32", tint32)


class Int64Coercer(ExprCoercer):
    @property
    def str_t(self):
        return 'int64'

    def _requires_conversion(self, t: HailType) -> bool:
        return t != tint64

    def can_coerce(self, t: HailType) -> bool:
        return t in (tbool, tint32, tint64)

    def _coerce(self, x):
        return x._method("toInt64", tint64)


class Float32Coercer(ExprCoercer):
    @property
    def str_t(self):
        return 'float32'

    def _requires_conversion(self, t: HailType) -> bool:
        return t != tfloat32

    def can_coerce(self, t: HailType) -> bool:
        return t in (tbool, tint32, tint64, tfloat32)

    def _coerce(self, x):
        return x._method("toFloat32", tfloat32)


class Float64Coercer(ExprCoercer):
    @property
    def str_t(self):
        return 'float64'

    def _requires_conversion(self, t: HailType) -> bool:
        return t != tfloat64

    def can_coerce(self, t: HailType) -> bool:
        return t in (tbool, tint32, tint64, tfloat32, tfloat64)

    def _coerce(self, x):
        return x._method("toFloat64", tfloat64)


class StringCoercer(ExprCoercer):
    @property
    def str_t(self):
        return 'str'

    def _requires_conversion(self, t: HailType) -> bool:
        return False

    def can_coerce(self, t: HailType) -> bool:
        return t == tstr


class CallCoercer(ExprCoercer):
    @property
    def str_t(self):
        return 'call'

    def _requires_conversion(self, t: HailType) -> bool:
        return False

    def can_coerce(self, t: HailType) -> bool:
        return t == tcall


class LocusCoercer(ExprCoercer):
    def __init__(self, rg: Optional['hl.ReferenceGenome'] = None):
        super(LocusCoercer, self).__init__()
        self.rg = rg

    @property
    def str_t(self):
        return f'locus<{"any" if not self.rg else self.rg}>'

    def _requires_conversion(self, t: HailType) -> bool:
        return False

    def can_coerce(self, t: HailType) -> bool:
        if self.rg:
            return t == tlocus(self.rg)
        else:
            return isinstance(t, tlocus)


class IntervalCoercer(ExprCoercer):
    def __init__(self, point_type: ExprCoercer = AnyCoercer()):
        super(IntervalCoercer, self).__init__()
        self.point_type = point_type

    @property
    def str_t(self):
        return f'interval<{self.point_type.str_t}>'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, tinterval)
        return self.point_type._requires_conversion(t.point_type)

    def can_coerce(self, t: HailType) -> bool:
        return isinstance(t, tinterval) and self.point_type.can_coerce(t.point_type)

    def _coerce(self, x):
        assert isinstance(x, hl.expr.IntervalExpression)
        return hl.interval(self.point_type.coerce(x.start),
                           self.point_type.coerce(x.end),
                           includes_start=x.includes_start,
                           includes_end=x.includes_end)


class ArrayCoercer(ExprCoercer):
    def __init__(self, ec: ExprCoercer = AnyCoercer()):
        super(ArrayCoercer, self).__init__()
        self.ec = ec

    @property
    def str_t(self):
        return f'array<{self.ec.str_t}>'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, tarray)
        return self.ec._requires_conversion(t.element_type)

    def can_coerce(self, t: HailType) -> bool:
        return isinstance(t, tarray) and self.ec.can_coerce(t.element_type)

    def _coerce(self, x: Expression):
        assert isinstance(x, hl.expr.ArrayExpression)
        return hl.map(lambda x_: self.ec.coerce(x_), x)


class NDArrayCoercer(ExprCoercer):
    def __init__(self, ec: ExprCoercer = AnyCoercer()):
        super(NDArrayCoercer, self).__init__()
        self.ec = ec

    @property
    def str_t(self):
        return f'ndarray<{self.ec.str_t}>'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, tndarray)
        return self.ec._requires_conversion(t.element_type)

    def can_coerce(self, t: HailType) -> bool:
        return isinstance(t, tndarray) and self.ec.can_coerce(t.element_type)

    def _coerce(self, x: Expression):
        assert isinstance(x, hl.expr.NDArrayExpression)
        return hl.map(lambda x_: self.ec.coerce(x_), x)


class SetCoercer(ExprCoercer):

    def __init__(self, ec: ExprCoercer = AnyCoercer()):
        super(SetCoercer, self).__init__()
        self.ec = ec

    @property
    def str_t(self):
        return f'set<{self.ec.str_t}>'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, tset)
        return self.ec._requires_conversion(t.element_type)

    def can_coerce(self, t: HailType) -> bool:
        return isinstance(t, tset) and self.ec.can_coerce(t.element_type)

    def _coerce(self, x: Expression):
        assert isinstance(x, hl.expr.SetExpression)
        return hl.map(lambda x_: self.ec.coerce(x_), x)


class DictCoercer(ExprCoercer):
    def __init__(self, kc: ExprCoercer = AnyCoercer(), vc: ExprCoercer = AnyCoercer()):
        super(DictCoercer, self).__init__()
        self.kc = kc
        self.vc = vc

    @property
    def str_t(self):
        return f'dict<{self.kc.str_t, self.vc.str_t}>'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, tdict)
        return self.kc._requires_conversion(t.key_type) or self.vc._requires_conversion(t.value_type)

    def can_coerce(self, t: HailType) -> bool:
        return isinstance(t, tdict) and self.kc.can_coerce(t.key_type) and self.vc.can_coerce(t.value_type)

    def _coerce(self, x: Expression):
        assert isinstance(x, hl.expr.DictExpression)
        if not self.kc._requires_conversion(x.dtype.key_type):
            # fast path
            return x.map_values(self.vc.coerce)
        else:
            return hl.dict(hl.map(lambda e: (self.kc.coerce(e[0]), self.vc.coerce(e[1])),
                                  hl.array(x)))


class TupleCoercer(ExprCoercer):
    def __init__(self, elements: Optional[List[ExprCoercer]] = None):
        super(TupleCoercer, self).__init__()
        self.elements = elements

    @property
    def str_t(self):
        if self.elements is None:
            return 'tuple'
        else:
            return f'tuple({", ".join(c.str_t for c in self.elements)})'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, ttuple)
        if self.elements is None:
            return False
        else:
            assert len(self.elements) == len(t.types)
            return any(c._requires_conversion(t_) for c, t_ in zip(self.elements, t.types))

    def can_coerce(self, t: HailType):
        if self.elements is None:
            return isinstance(t, ttuple)
        else:
            return (isinstance(t, ttuple)
                    and len(t.types) == len(self.elements)
                    and all(c.can_coerce(t_) for c, t_ in zip(self.elements, t.types)))

    def _coerce(self, x: Expression):
        assert isinstance(x, hl.expr.TupleExpression)
        return hl.tuple(c.coerce(e) for c, e in zip(self.elements, x))


class StructCoercer(ExprCoercer):
    def __init__(self, fields: Optional[Dict[str, ExprCoercer]] = None):
        super(StructCoercer, self).__init__()
        self.fields = fields

    @property
    def str_t(self) -> str:
        if self.fields is None:
            return 'struct'
        else:
            field_strs = ', '.join(f'{escape_parsable(name)}: {c.str_t}' for name, c in self.fields.items())
            return f'struct{{{field_strs}}})'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, tstruct)
        if self.fields is None:
            return False
        else:
            return any(c._requires_conversion(t[name]) for name, c in self.fields.items())

    def can_coerce(self, t: HailType):
        if self.fields is None:
            return isinstance(t, tstruct)
        else:
            return (isinstance(t, tstruct)
                    and len(t) == len(self.fields)
                    and all(expected[0] == actual[0] and expected[1].can_coerce(actual[1])
                            for expected, actual in zip(self.fields.items(), t.items())))

    def _coerce(self, x: Expression):
        assert isinstance(x, hl.expr.StructExpression)
        assert list(x.keys()) == list(self.fields.keys())
        return hl.struct(**{name: c.coerce(x[name]) for name, c in self.fields.items()})


class UnionCoercer(ExprCoercer):
    def __init__(self, cases: Optional[Dict[str, ExprCoercer]] = None):
        super(UnionCoercer, self).__init__()
        self.cases = cases

    @property
    def str_t(self) -> str:
        if self.cases is None:
            return 'union'
        else:
            case_strs = ', '.join(f'{escape_parsable(name)}: {c.str_t}' for name, c in self.cases.items())
            return f'union{{{case_strs}}})'

    def _requires_conversion(self, t: HailType) -> bool:
        assert isinstance(t, tunion)
        if self.cases is None:
            return False
        else:
            return any(c._requires_conversion(t[name]) for name, c in self.cases.items())

    def can_coerce(self, t: HailType):
        if self.cases is None:
            return isinstance(t, tunion)
        else:
            return (isinstance(t, tunion)
                    and len(t) == len(self.cases)
                    and all(expected[0] == actual[0] and expected[1].can_coerce(actual[1])
                            for expected, actual in zip(self.cases.items(), t.items())))

    def _coerce(self, x: Expression):
        assert isinstance(x, hl.expr.StructExpression)
        assert list(x.keys()) == list(self.cases.keys())
        raise NotImplementedError()


class OneOfExprCoercer(ExprCoercer):
    def __init__(self, *options: ExprCoercer):
        super(OneOfExprCoercer, self).__init__()
        options_ = []
        for o in options:
            if isinstance(o, OneOfExprCoercer):
                options_.extend(o.options)
            else:
                options_.append(o)
        self.options = options_

    @property
    def str_t(self) -> str:
        return ' or '.join(o.str_t for o in self.options)

    def _requires_conversion(self, t: HailType) -> bool:
        return all(o._requires_conversion(t) for o in filter(lambda c: c.can_coerce(t), self.options))

    def can_coerce(self, t: HailType) -> bool:
        return any(o.can_coerce(t) for o in self.options)

    def _coerce(self, x: Expression) -> Expression:
        first_coercer = next(filter(lambda o: o.can_coerce(x.dtype), self.options))
        return first_coercer._coerce(x)


expr_any = AnyCoercer()
expr_oneof = OneOfExprCoercer
expr_int32 = Int32Coercer()
expr_int64 = Int64Coercer()
expr_float32 = Float32Coercer()
expr_float64 = Float64Coercer()
expr_call = CallCoercer()
expr_str = StringCoercer()
expr_bool = BoolCoercer()
expr_locus = LocusCoercer
expr_interval = IntervalCoercer
expr_array = ArrayCoercer
expr_ndarray = NDArrayCoercer
expr_set = SetCoercer
expr_dict = DictCoercer
expr_tuple = TupleCoercer
expr_struct = StructCoercer
expr_union = UnionCoercer
expr_numeric = expr_oneof(expr_int32, expr_int64, expr_float32, expr_float64)

primitives: Dict[HailType, ExprCoercer] = {
    tint32: expr_int32,
    tint64: expr_int64,
    tfloat32: expr_float32,
    tfloat64: expr_float64,
    tbool: expr_bool,
    tcall: expr_call,
    tstr: expr_str
}


def coercer_from_dtype(t: HailType) -> ExprCoercer:
    if t in primitives:
        return primitives[t]
    elif isinstance(t, tlocus):
        return expr_locus(t.reference_genome)
    elif isinstance(t, tinterval):
        return expr_interval(coercer_from_dtype(t.point_type))
    elif isinstance(t, tarray):
        return expr_array(coercer_from_dtype(t.element_type))
    elif isinstance(t, tndarray):
        return expr_ndarray(coercer_from_dtype(t.element_type))
    elif isinstance(t, tset):
        return expr_set(coercer_from_dtype(t.element_type))
    elif isinstance(t, tdict):
        return expr_dict(coercer_from_dtype(t.key_type),
                         coercer_from_dtype(t.value_type))
    elif isinstance(t, ttuple):
        return expr_tuple([coercer_from_dtype(t_) for t_ in t.types])
    elif isinstance(t, tstruct):
        return expr_struct({name: coercer_from_dtype(t_) for name, t_ in t.items()})
    else:
        assert isinstance(t, tunion)
        return expr_union({name: coercer_from_dtype(t_) for name, t_ in t.items()})



import re
import json
from subprocess import call, Popen, PIPE


def init_parser(parser):
    parser.add_argument('name', type=str, help='Cluster name.')
    parser.add_argument('--dest', '-d', required=True, type=str, help="Directory for diagnose output -- must be local.")
    parser.add_argument('--hail-log', '-l', required=False, type=str, default='/home/hail/hail.log',
                        help="Path for hail.log file.")
    parser.add_argument('--overwrite', required=False, action='store_true',
                        help="Delete dest directory before adding new files.")
    parser.add_argument('--no-diagnose', required=False, action='store_true',
                        help="Do not run gcloud dataproc clusters diagnose.")
    parser.add_argument('--compress', '-z', required=False, action='store_true', help="GZIP all files.")
    parser.add_argument('--workers', required=False, nargs='*', help="Specific workers to get log files from.")
    parser.add_argument('--take', required=False, type=int, default=None,
                        help="Only download logs from the first N workers.")


def main(args, pass_through_args):  # pylint: disable=unused-argument
    print("Diagnosing cluster '{}'...".format(args.name))

    is_local = not args.dest.startswith("gs://")

    if args.overwrite:
        if is_local:
            call('rm -r {dir}'.format(dir=args.dest), shell=True)
        else:
            call('gsutil -m rm -r {dir}'.format(dir=args.dest), shell=True)

    master_dest = args.dest.rstrip('/') + "/master/"
    worker_dest = args.dest.rstrip('/') + "/workers/"

    if is_local:
        call('mkdir -p {dir}'.format(dir=master_dest), shell=True)
        call('mkdir -p {dir}'.format(dir=worker_dest), shell=True)

    desc = json.loads(Popen('gcloud dataproc clusters describe {name} --format json'.format(name=args.name),
                            shell=True,
                            stdout=PIPE,
                            stderr=PIPE).communicate()[0].strip())

    config = desc['config']

    master = config['masterConfig']['instanceNames'][0]
    try:
        workers = config['workerConfig']['instanceNames'] + config['secondaryWorkerConfig']['instanceNames']
    except KeyError:
        workers = config['workerConfig']['instanceNames']
    zone = re.search(r'zones/(?P<zone>\S+)$', config['gceClusterConfig']['zoneUri']).group('zone')

    if args.workers:
        invalid_workers = set(args.workers).difference(set(workers))
        assert len(invalid_workers) == 0, "Non-existent workers specified: " + ", ".join(invalid_workers)
        workers = args.workers

    if args.take:
        assert args.take > 0 and args.take <= len(
            workers), "Number of workers to take must be in the range of [0, nWorkers]. Found " + args.take + "."
        workers = workers[:args.take]

    def gcloud_ssh(remote, command):
        return 'gcloud compute ssh {remote} --zone {zone} --command "{command}"'.format(remote=remote, zone=zone,
                                                                                        command=command)

    def gcloud_copy_files(remote, src, dest):
        return 'gcloud compute copy-files {remote}:{src} {dest} --zone {zone}'.format(remote=remote, src=src, dest=dest,
                                                                                      zone=zone)

    def gsutil_cp(src, dest):
        return 'gsutil -m cp -r {src} {dest}'.format(src=src, dest=dest)

    def copy_files_tmp(remote, files, dest, tmp):
        init_cmd = ['mkdir -p {tmp}; rm -r {tmp}/*'.format(tmp=tmp)]

        copy_tmp_cmds = ['sudo cp -r {file} {tmp}'.format(file=file, tmp=tmp) for file in files]
        copy_tmp_cmds.append('sudo chmod -R 777 {tmp}'.format(tmp=tmp))

        if args.compress:
            copy_tmp_cmds.append('sudo find ' + tmp + ' -type f ! -name \'*.gz\' -exec gzip "{}" \\;')

        call(gcloud_ssh(remote, '; '.join(init_cmd + copy_tmp_cmds)), shell=True)

        if not is_local:
            copy_dest_cmd = gcloud_ssh(remote, 'gsutil -m cp -r {tmp} {dest}'.format(tmp=tmp, dest=dest))
        else:
            copy_dest_cmd = gcloud_copy_files(remote, tmp, dest)

        call(copy_dest_cmd, shell=True)

    if not args.no_diagnose:
        diagnose_tar_path = re.search(r'Diagnostic results saved in: (?P<tarfile>gs://\S+diagnostic\.tar)',
                                      str(Popen('gcloud dataproc clusters diagnose {name}'.format(name=args.name),
                                                shell=True,
                                                stdout=PIPE,
                                                stderr=PIPE).communicate())).group('tarfile')

        call(gsutil_cp(diagnose_tar_path, args.dest), shell=True)

    master_log_files = ['/var/log/hive/hive-*',
                        '/var/log/google-dataproc-agent.0.log',
                        '/var/log/dataproc-initialization-script-0.log',
                        '/var/log/hadoop-mapreduce/mapred-mapred-historyserver*',
                        '/var/log/hadoop-hdfs/*-m.*',
                        '/var/log/hadoop-yarn/yarn-yarn-resourcemanager-*-m.*',
                        args.hail_log
                        ]

    copy_files_tmp(master, master_log_files, master_dest, '/tmp/' + master + '/')

    worker_log_files = ['/var/log/hadoop-hdfs/hadoop-hdfs-datanode-*.*',
                        '/var/log/dataproc-startup-script.log',
                        '/var/log/hadoop-yarn/yarn-yarn-nodemanager-*.*']

    for worker in workers:
        copy_files_tmp(worker, worker_log_files, worker_dest, '/tmp/' + worker + '/')
        copy_files_tmp(worker, ['/var/log/hadoop-yarn/userlogs/'], args.dest, '/tmp/hadoop-yarn/')



from typing import List, Optional

from hail import MatrixTable
from hail.linalg import BlockMatrix
from hail.ir import MatrixMultiWrite, MatrixNativeMultiWriter, BlockMatrixMultiWrite, BlockMatrixBinaryMultiWriter, BlockMatrixTextMultiWriter
from hail.typecheck import nullable, sequenceof, typecheck, enumeration
from hail.utils.java import Env

@typecheck(mts=sequenceof(MatrixTable),
           prefix=str,
           overwrite=bool,
           stage_locally=bool)
def write_matrix_tables(mts: List[MatrixTable], prefix: str, overwrite: bool = False,
                        stage_locally: bool = False):
    writer = MatrixNativeMultiWriter(prefix, overwrite, stage_locally)
    Env.backend().execute(MatrixMultiWrite([mt._mir for mt in mts], writer))

@typecheck(bms=sequenceof(BlockMatrix),
           prefix=str,
           overwrite=bool)
def block_matrices_tofiles(bms: List[BlockMatrix], prefix: str, overwrite: bool = False):
    writer = BlockMatrixBinaryMultiWriter(prefix, overwrite)
    Env.backend().execute(BlockMatrixMultiWrite([bm._bmir for bm in bms], writer))

@typecheck(bms=sequenceof(BlockMatrix),
           prefix=str,
           overwrite=bool,
           delimiter=str,
           header=nullable(str),
           add_index=bool,
           compression=nullable(enumeration('gz', 'bgz')),
           custom_filenames=nullable(sequenceof(str)))
def export_block_matrices(bms: List[BlockMatrix], prefix: str, overwrite: bool = False,
                          delimiter: str = '\t', header: Optional[str] = None,  add_index: bool = False,
                          compression: Optional[str] = None, custom_filenames=None):

    if custom_filenames:
        assert len(custom_filenames) == len(bms), "Number of block matrices and number of custom filenames must be equal"

    writer = BlockMatrixTextMultiWriter(prefix, overwrite, delimiter, header, add_index, compression, custom_filenames)
    Env.backend().execute(BlockMatrixMultiWrite([bm._bmir for bm in bms], writer))



from .base_expression import *
from .typed_expressions import *
from .expression_typecheck import *
from .expression_utils import *

__all__ = ['Indices',
           'Aggregation',
           'apply_expr',
           'construct_expr',
           'construct_variable',
           'construct_reference',
           'cast',
           'impute_type',
           'to_expr',
           'cast_expr',
           'unify_all',
           'unify_types_limited',
           'unify_types',
           'unify_exprs',
           'Expression',
           'ExpressionException',
           'Expression',
           'ArrayExpression',
           'ArrayNumericExpression',
           'BooleanExpression',
           'CallExpression',
           'CollectionExpression',
           'DictExpression',
           'IntervalExpression',
           'LocusExpression',
           'NumericExpression',
           'Int32Expression',
           'Int64Expression',
           'Float32Expression',
           'Float64Expression',
           'SetExpression',
           'StringExpression',
           'StructExpression',
           'TupleExpression',
           'NDArrayExpression',
           'NDArrayNumericExpression',
           'analyze',
           'check_entry_indexed',
           'check_row_indexed',
           'get_refs',
           'extract_refs_by_indices',
           'eval',
           'eval_typed',
           'eval_timed',
           'expr_any',
           'expr_int32',
           'expr_int64',
           'expr_float32',
           'expr_float64',
           'expr_bool',
           'expr_str',
           'expr_call',
           'expr_locus',
           'expr_struct',
           'expr_numeric',
           'expr_array',
           'expr_set',
           'expr_dict',
           'expr_struct',
           'expr_tuple',
           'expr_interval',
           'expr_oneof',
           'matrix_table_source',
           'table_source',
           ]



import sys

import argparse

from . import config
from . import deploy


def parser():
    main_parser = argparse.ArgumentParser(
        prog='hailctl dev',
        description='Manage Hail development utilities.')
    subparsers = main_parser.add_subparsers()

    config_parser = subparsers.add_parser(
        'config',
        help='Configure deployment',
        description='Configure deployment')

    config.cli.init_parser(config_parser)

    deploy_parser = subparsers.add_parser(
        'deploy',
        help='Deploy a branch',
        description='Deploy a branch')

    deploy.cli.init_parser(deploy_parser)

    return main_parser


def main(args):
    p = parser()

    if not args:
        p.print_help()
        sys.exit(0)
    else:
        module = args[0]
        if module == 'deploy':
            from .deploy import cli
            args, _ = p.parse_known_args(args=args)
            cli.main(args)
        elif module == 'config':
            from .config import cli
            args, _ = p.parse_known_args(args=args)
            cli.main(args)
        elif module in ('-h', '--help', 'help'):
            p.print_help()
        else:
            sys.stderr.write(f"ERROR: no such module: {module!r}")
            p.print_help()
            sys.exit(1)



import os.path
import sys
from subprocess import check_call
import yaml
import pkg_resources


def init_parser(parser):
    parser.add_argument('name', type=str, help='Cluster name.')
    parser.add_argument('--wheel', type=str, help='New Hail installation.')
    parser.add_argument('--num-workers', '--n-workers', '-w', type=int,
                        help='New number of worker machines (min. 2).')
    parser.add_argument('--num-preemptible-workers', '--n-pre-workers', '-p', type=int,
                        help='New number of preemptible worker machines.')
    parser.add_argument('--graceful-decommission-timeout', '--graceful', type=str,
                        help='If set, cluster size downgrade will use graceful decommissioning with the given timeout (e.g. "60m").')
    parser.add_argument('--max-idle', type=str, help='New maximum idle time before shutdown (e.g. "60m").')
    parser.add_argument('--dry-run', action='store_true', help="Print gcloud dataproc command, but don't run it.")
    parser.add_argument('--zone', '-z', default='us-central1-b', type=str,
                        help='Compute zone for Dataproc cluster (default: %(default)s).')
    parser.add_argument('--update-hail-version', action='store_true', help="Update the version of hail running on cluster to match "
                        "the currently installed version.")


def main(args, pass_through_args):
    modify_args = []
    if args.num_workers is not None:
        modify_args.append('--num-workers={}'.format(args.num_workers))

    if args.num_preemptible_workers is not None:
        modify_args.append('--num-preemptible-workers={}'.format(args.num_preemptible_workers))

    if args.graceful_decommission_timeout:
        if not modify_args:
            sys.exit("Error: Cannot use --graceful-decommission-timeout without resizing the cluster.")
        modify_args.append('--graceful-decommission-timeout={}'.format(args.graceful_decommission_timeout))

    if args.max_idle:
        modify_args.append('--max-idle={}'.format(args.max_idle))

    if args.update_hail_version and args.wheel:
        sys.exit("Error: Cannot specify both --update-hail-version and --wheel")

    if modify_args:
        cmd = ['gcloud',
               'dataproc',
               'clusters',
               'update',
               args.name] + modify_args

        if args.beta:
            cmd.insert(1, 'beta')

        cmd.extend(pass_through_args)

        # print underlying gcloud command
        print(' '.join(cmd[:5]) + ' \\\n    ' + ' \\\n    '.join(cmd[5:]))

        # Update cluster
        if not args.dry_run:
            print("Updating cluster '{}'...".format(args.name))
            check_call(cmd)

    wheel = None
    if args.update_hail_version:
        assert pkg_resources.resource_exists('hailtop.hailctl', "deploy.yaml")
        updated_wheel = yaml.safe_load(
            pkg_resources.resource_stream('hailtop.hailctl', "deploy.yaml"))['dataproc']['wheel']
        wheel = updated_wheel
    else:
        wheel = args.wheel

    if wheel is not None:
        wheelfile = os.path.basename(wheel)
        cmds = []
        if wheel.startswith("gs://"):
            cmds.append([
                'gcloud',
                'compute',
                'ssh',
                '{}-m'.format(args.name),
                '--zone={}'.format(args.zone),
                '--',
                f'sudo gsutil cp {wheel} /tmp/ && '
                'sudo /opt/conda/default/bin/pip uninstall -y hail && '
                f'sudo /opt/conda/default/bin/pip install --no-dependencies /tmp/{wheelfile}'
            ])
        else:
            cmds.extend([
                [
                    'gcloud',
                    'compute',
                    'scp',
                    '--zone={}'.format(args.zone),
                    wheel,
                    '{}-m:/tmp/'.format(args.name)
                ],
                [
                    'gcloud',
                    'compute',
                    'ssh',
                    f'{args.name}-m',
                    f'--zone={args.zone}',
                    '--',
                    'sudo /opt/conda/default/bin/pip uninstall -y hail && '
                    f'sudo /opt/conda/default/bin/pip install --no-dependencies /tmp/{wheelfile}'
                ]
            ])

        for cmd in cmds:
            print(cmd)
            check_call(cmd)

    if not wheel and not modify_args and pass_through_args:
        sys.stderr.write('ERROR: found pass-through arguments but not known modification args.')
        sys.exit(1)



import asyncio


class CalledProcessError(Exception):
    def __init__(self, command, returncode, outerr):
        super().__init__()
        self.command = command
        self.returncode = returncode
        self.outerr = outerr

    def __str__(self):
        return (f'Command {self.command} returned non-zero exit status {self.returncode}.'
                f' Output:\n{self.outerr}')


async def check_shell_output(script):
    proc = await asyncio.create_subprocess_exec(
        '/bin/bash', '-c', script,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)
    outerr = await proc.communicate()
    if proc.returncode != 0:
        raise CalledProcessError(script, proc.returncode, outerr)
    return outerr


async def check_shell(script):
    # discard output
    await check_shell_output(script)



from hail import call
from hail.expr.expressions import expr_array, expr_call, expr_int32
from hail.typecheck import typecheck


@typecheck(lgt=expr_call, la=expr_array(expr_int32))
def lgt_to_gt(lgt, la):
    """Transform LGT into GT using local alleles array.

    Parameters
    ----------
    lgt : :class:`.CallExpression`
        LGT value.
    la : :class:`.ArrayExpression`
        Local alleles array.

    Returns
    -------
    :class:`.CallExpression`

    Notes
    -----
    This function assumes diploid genotypes.
    """
    return call(la[lgt[0]], la[lgt[1]])



import pkg_resources
from pyspark import SparkContext

import hail
from hail.genetics.reference_genome import ReferenceGenome
from hail.typecheck import nullable, typecheck, typecheck_method, enumeration, dictof
from hail.utils import get_env_or_default
from hail.utils.java import Env, joption, FatalError, connect_logger, install_exception_handler, uninstall_exception_handler, warn
from hail.backend import Backend, ServiceBackend, SparkBackend

import sys
import os


class HailContext(object):
    @typecheck_method(sc=nullable(SparkContext),
                      app_name=str,
                      master=nullable(str),
                      local=str,
                      log=nullable(str),
                      quiet=bool,
                      append=bool,
                      min_block_size=int,
                      branching_factor=int,
                      tmp_dir=nullable(str),
                      default_reference=str,
                      idempotent=bool,
                      global_seed=nullable(int),
                      spark_conf=nullable(dictof(str, str)),
                      optimizer_iterations=nullable(int),
                      _backend=nullable(Backend))
    def __init__(self, sc=None, app_name="Hail", master=None, local='local[*]',
                 log=None, quiet=False, append=False,
                 min_block_size=1, branching_factor=50, tmp_dir=None,
                 default_reference="GRCh37", idempotent=False,
                 global_seed=6348563392232659379, spark_conf=None,
                 optimizer_iterations=None, _backend=None):

        if Env._hc:
            if idempotent:
                return
            else:
                warn('Hail has already been initialized. If this call was intended to change configuration,'
                     ' close the session with hl.stop() first.')

        py_version = version()
        
        if log is None:
            log = hail.utils.timestamp_path(os.path.join(os.getcwd(), 'hail'),
                                            suffix=f'-{py_version}.log')
        self._log = log

        tmp_dir = get_env_or_default(tmp_dir, 'TMPDIR', '/tmp')
        self.tmp_dir = tmp_dir

        optimizer_iterations = get_env_or_default(optimizer_iterations, 'HAIL_OPTIMIZER_ITERATIONS', 3)

        if _backend is None:
            if os.environ.get('HAIL_APISERVER_URL') is not None:
                _backend = ServiceBackend()
            else:
                _backend = SparkBackend(
                    idempotent, sc, spark_conf, app_name, master, local, log,
                    quiet, append, min_block_size, branching_factor, tmp_dir,
                    optimizer_iterations)
        self._backend = _backend

        self._warn_cols_order = True
        self._warn_entries_order = True

        super(HailContext, self).__init__()

        # do this at the end in case something errors, so we don't raise the above error without a real HC
        Env._hc = self

        ReferenceGenome._from_config(_backend.get_reference('GRCh37'), True)
        ReferenceGenome._from_config(_backend.get_reference('GRCh38'), True)
        ReferenceGenome._from_config(_backend.get_reference('GRCm38'), True)
        ReferenceGenome._from_config(_backend.get_reference('CanFam3'), True)

        if default_reference in ReferenceGenome._references:
            self._default_ref = ReferenceGenome._references[default_reference]
        else:
            self._default_ref = ReferenceGenome.read(default_reference)

        if not quiet:
            sys.stderr.write(
                'Welcome to\n'
                '     __  __     <>__\n'
                '    / /_/ /__  __/ /\n'
                '   / __  / _ `/ / /\n'
                '  /_/ /_/\\_,_/_/_/   version {}\n'.format(py_version))

            if py_version.startswith('devel'):
                sys.stderr.write('NOTE: This is a beta version. Interfaces may change\n'
                                 '  during the beta period. We recommend pulling\n'
                                 '  the latest changes weekly.\n')
            sys.stderr.write(f'LOGGING: writing to {log}\n')

        install_exception_handler()
        Env.set_seed(global_seed)


    @property
    def default_reference(self):
        return self._default_ref

    def stop(self):
        self._backend.stop()
        self._backend = None
        Env._jvm = None
        Env._hc = None
        uninstall_exception_handler()
        Env._dummy_table = None
        Env._seed_generator = None
        hail.ir.clear_session_functions()
        ReferenceGenome._references = {}


@typecheck(sc=nullable(SparkContext),
           app_name=str,
           master=nullable(str),
           local=str,
           log=nullable(str),
           quiet=bool,
           append=bool,
           min_block_size=int,
           branching_factor=int,
           tmp_dir=str,
           default_reference=enumeration('GRCh37', 'GRCh38', 'GRCm38', 'CanFam3'),
           idempotent=bool,
           global_seed=nullable(int),
           spark_conf=nullable(dictof(str, str)),
           _optimizer_iterations=nullable(int),
           _backend=nullable(Backend))
def init(sc=None, app_name='Hail', master=None, local='local[*]',
         log=None, quiet=False, append=False,
         min_block_size=0, branching_factor=50, tmp_dir='/tmp',
         default_reference='GRCh37', idempotent=False,
         global_seed=6348563392232659379,
         spark_conf=None,
         _optimizer_iterations=None,
         _backend=None):
    """Initialize Hail and Spark.

    Examples
    --------
    Import and initialize Hail using GRCh38 as the default reference genome:

    >>> import hail as hl
    >>> hl.init(default_reference='GRCh38')  # doctest: +SKIP

    Notes
    -----
    Hail is not only a Python library; most of Hail is written in Java/Scala
    and runs together with Apache Spark in the Java Virtual Machine (JVM).
    In order to use Hail, a JVM needs to run as well. The :func:`.init`
    function is used to initialize Hail and Spark.

    This function also sets global configuration parameters used for the Hail
    session, like the default reference genome and log file location.

    This function will be called automatically (with default parameters) if
    any Hail functionality requiring the backend (most of the libary!) is used.
    To initialize Hail explicitly with non-default arguments, be sure to do so
    directly after importing the module, as in the above example.

    Note
    ----
    If a :class:`pyspark.SparkContext` is already running, then Hail must be
    initialized with it as an argument:

    >>> hl.init(sc=sc)  # doctest: +SKIP

    See Also
    --------
    :func:`.stop`

    Parameters
    ----------
    sc : pyspark.SparkContext, optional
        Spark context. By default, a Spark context will be created.
    app_name : :obj:`str`
        Spark application name.
    master : :obj:`str`, optional
        Spark master.
    local : :obj:`str`
       Local-mode master, used if `master` is not defined here or in the
       Spark configuration.
    log : :obj:`str`
        Local path for Hail log file. Does not currently support distributed
        file systems like Google Storage, S3, or HDFS.
    quiet : :obj:`bool`
        Print fewer log messages.
    append : :obj:`bool`
        Append to the end of the log file.
    min_block_size : :obj:`int`
        Minimum file block size in MB.
    branching_factor : :obj:`int`
        Branching factor for tree aggregation.
    tmp_dir : :obj:`str`
        Temporary directory for Hail files. Must be a network-visible
        file path.
    default_reference : :obj:`str`
        Default reference genome. Either ``'GRCh37'``, ``'GRCh38'``,
        ``'GRCm38'``, or ``'CanFam3'``.
    idempotent : :obj:`bool`
        If ``True``, calling this function is a no-op if Hail has already been initialized.
    global_seed : :obj:`int`, optional
        Global random seed.
    spark_conf : :obj:`dict[str, str]`, optional
        Spark configuration parameters.
    """
    HailContext(sc, app_name, master, local, log, quiet, append,
                min_block_size, branching_factor, tmp_dir,
                default_reference, idempotent, global_seed, spark_conf,
                _optimizer_iterations,_backend)


def version():
    """Get the installed hail version.

    Returns
    -------
    str
    """
    if hail.__version__ is None:
        # https://stackoverflow.com/questions/6028000/how-to-read-a-static-file-from-inside-a-python-package
        hail.__version__ = pkg_resources.resource_string(__name__, 'hail_version').decode().strip()
    return hail.__version__


def _hail_cite_url():
    v = version()
    [tag, sha_prefix] = v.split("-")
    if pkg_resources.resource_exists(__name__, "hail-all-spark.jar"):
        # pip installed
        return f"https://github.com/hail-is/hail/releases/tag/{tag}"
    return f"https://github.com/hail-is/hail/commit/{sha_prefix}"


def citation(*, bibtex=False):
    """Generate a Hail citation.

    Parameters
    ----------
    bibtex : bool
        Generate a citation in BibTeX form.

    Returns
    -------
    str
    """
    if bibtex:
        return f"@misc{{Hail," \
            f"  author = {{Hail Team}}," \
            f"  title = {{Hail}}," \
            f"  howpublished = {{\\url{{{_hail_cite_url()}}}}}" \
            f"}}"
    return f"Hail Team. Hail {version()}. {_hail_cite_url()}."


def cite_hail():
    return citation(bibtex=False)


def cite_hail_bibtex():
    return citation(bibtex=True)


def stop():
    """Stop the currently running Hail session."""
    if Env._hc:
        Env.hc().stop()

def spark_context():
    """Returns the active Spark context.

    Returns
    -------
    :class:`pyspark.SparkContext`
    """
    return Env.backend().sc

def current_backend():
    return Env.hc()._backend

def default_reference():
    """Returns the default reference genome ``'GRCh37'``.

    Returns
    -------
    :class:`.ReferenceGenome`
    """
    return Env.hc().default_reference

def get_reference(name) -> 'hail.ReferenceGenome':
    """Returns the reference genome corresponding to `name`.

    Notes
    -----

    Hail's built-in references are ``'GRCh37'``, ``GRCh38'``, ``'GRCm38'``, and
    ``'CanFam3'``.
    The contig names and lengths come from the GATK resource bundle:
    `human_g1k_v37.dict
    <ftp://gsapubftp-anonymous@ftp.broadinstitute.org/bundle/b37/human_g1k_v37.dict>`__
    and `Homo_sapiens_assembly38.dict
    <ftp://gsapubftp-anonymous@ftp.broadinstitute.org/bundle/hg38/Homo_sapiens_assembly38.dict>`__.


    If ``name='default'``, the value of :func:`.default_reference` is returned.

    Parameters
    ----------
    name : :obj:`str`
        Name of a previously loaded reference genome or one of Hail's built-in
        references: ``'GRCh37'``, ``'GRCh38'``, ``'GRCm38'``, ``'CanFam3'``, and
        ``'default'``.

    Returns
    -------
    :class:`.ReferenceGenome`
    """
    Env.hc()
    if name == 'default':
        return default_reference()
    else:
        return ReferenceGenome._references[name]


@typecheck(seed=int)
def set_global_seed(seed):
    """Sets Hail's global seed to `seed`.

    Parameters
    ----------
    seed : :obj:`int`
        Integer used to seed Hail's random number generator
    """

    Env.set_seed(seed)



def _set_flags(**flags):
    available = set(Env.backend()._jhc.flags().available())
    invalid = []
    for flag, value in flags.items():
        if flag in available:
            Env.backend()._jhc.flags().set(flag, value)
        else:
            invalid.append(flag)
    if len(invalid) != 0:
        raise FatalError("Flags {} not valid. Valid flags: \n    {}"
                         .format(', '.join(invalid), '\n    '.join(available)))


def _get_flags(*flags):
    return {flag: Env.backend()._jhc.flags().get(flag) for flag in flags}


def debug_info():
    hail_jar_path = None
    if pkg_resources.resource_exists(__name__, "hail-all-spark.jar"):
        hail_jar_path = pkg_resources.resource_filename(__name__, "hail-all-spark.jar")
    return {
        'spark_conf': spark_context()._conf.getAll(),
        'hail_jar_path': hail_jar_path,
        'version': version()
    }



import hail as hl
from .java import FatalError, Env, info
from .misc import local_path_uri, new_local_temp_dir
import os
import zipfile
from urllib.request import urlretrieve
from hailtop.utils import sync_retry_transient_errors

__all__ = [
    'get_1kg',
    'get_movie_lens'
]

resources = {
    '1kg_annotations': 'https://storage.googleapis.com/hail-tutorial/1kg_annotations.txt',
    '1kg_matrix_table': 'https://storage.googleapis.com/hail-tutorial/1kg.vcf.bgz',
    'ensembl_gene_annotations': 'https://storage.googleapis.com/hail-tutorial/ensembl_gene_annotations.txt',
    'movie_lens_100k': 'http://files.grouplens.org/datasets/movielens/ml-100k.zip',
}

tmp_dir: str = None


def init_temp_dir():
    global tmp_dir
    if tmp_dir is None:
        tmp_dir = new_local_temp_dir()


def _mkdir(jhc, path):
    if not Env.jutils().dirExists(jhc, path):
        Env.jutils().mkdir(jhc, path)


def get_1kg(output_dir, overwrite: bool = False):
    """Download subset of the `1000 Genomes <http://www.internationalgenome.org/>`__
    dataset and sample annotations.

    Notes
    -----
    The download is about 15M.

    Parameters
    ----------
    output_dir
        Directory in which to write data.
    overwrite
        If ``True``, overwrite any existing files/directories at `output_dir`.
    """
    jhc = Env.backend()._jhc

    _mkdir(jhc, output_dir)

    matrix_table_path = os.path.join(output_dir, '1kg.mt')
    vcf_path = os.path.join(output_dir, '1kg.vcf.bgz')
    sample_annotations_path = os.path.join(output_dir, '1kg_annotations.txt')
    gene_annotations_path = os.path.join(output_dir, 'ensembl_gene_annotations.txt')

    if (overwrite
            or not Env.jutils().dirExists(jhc, matrix_table_path)
            or not Env.jutils().fileExists(jhc, sample_annotations_path)
            or not Env.jutils().fileExists(jhc, vcf_path)
            or not Env.jutils().fileExists(jhc, gene_annotations_path)):
        init_temp_dir()
        tmp_vcf = os.path.join(tmp_dir, '1kg.vcf.bgz')
        source = resources['1kg_matrix_table']
        info(f'downloading 1KG VCF ...\n'
             f'  Source: {source}')
        sync_retry_transient_errors(urlretrieve, resources['1kg_matrix_table'], tmp_vcf)
        cluster_readable_vcf = Env.jutils().copyToTmp(jhc, local_path_uri(tmp_vcf), 'vcf')
        info('importing VCF and writing to matrix table...')
        hl.import_vcf(cluster_readable_vcf, min_partitions=16).write(matrix_table_path, overwrite=True)

        tmp_sample_annot = os.path.join(tmp_dir, '1kg_annotations.txt')
        source = resources['1kg_annotations']
        info(f'downloading 1KG annotations ...\n'
             f'  Source: {source}')
        sync_retry_transient_errors(urlretrieve, source, tmp_sample_annot)

        tmp_gene_annot = os.path.join(tmp_dir, 'ensembl_gene_annotations.txt')
        source = resources['ensembl_gene_annotations']
        info(f'downloading Ensembl gene annotations ...\n'
             f'  Source: {source}')
        sync_retry_transient_errors(urlretrieve, source, tmp_gene_annot)

        hl.hadoop_copy(local_path_uri(tmp_sample_annot), sample_annotations_path)
        hl.hadoop_copy(local_path_uri(tmp_gene_annot), gene_annotations_path)
        hl.hadoop_copy(local_path_uri(tmp_vcf), vcf_path)
        info('Done!')
    else:
        info('1KG files found')


def get_movie_lens(output_dir, overwrite: bool = False):
    """Download public Movie Lens dataset.

    Notes
    -----
    The download is about 6M.

    See the
    `MovieLens website <https://grouplens.org/datasets/movielens/100k/>`__
    for more information about this dataset.

    Parameters
    ----------
    output_dir
        Directory in which to write data.
    overwrite
        If ``True``, overwrite existing files/directories at those locations.
    """

    jhc = Env.backend()._jhc

    _mkdir(jhc, output_dir)

    paths = [os.path.join(output_dir, x) for x in ['movies.ht', 'ratings.ht', 'users.ht']]
    if overwrite or any(not Env.jutils().dirExists(jhc, f) for f in paths):
        init_temp_dir()
        source = resources['movie_lens_100k']
        tmp_path = os.path.join(tmp_dir, 'ml-100k.zip')
        info(f'downloading MovieLens-100k data ...\n'
             f'  Source: {source}')
        sync_retry_transient_errors(urlretrieve, source, tmp_path)
        with zipfile.ZipFile(tmp_path, 'r') as z:
            z.extractall(tmp_dir)

        user_table_path = os.path.join(os.path.join(tmp_dir, 'ml-100k', 'u.user'))
        movie_table_path = os.path.join(os.path.join(tmp_dir, 'ml-100k', 'u.item'))
        ratings_table_path = os.path.join(os.path.join(tmp_dir, 'ml-100k', 'u.data'))
        assert (os.path.exists(user_table_path))
        assert (os.path.exists(movie_table_path))
        assert (os.path.exists(ratings_table_path))

        user_cluster_readable = Env.jutils().copyToTmp(jhc, local_path_uri(user_table_path), 'txt')
        movie_cluster_readable = Env.jutils().copyToTmp(jhc, local_path_uri(movie_table_path), 'txt')
        ratings_cluster_readable = Env.jutils().copyToTmp(jhc, local_path_uri(ratings_table_path), 'txt')

        [movies_path, ratings_path, users_path] = paths

        genres = ['Action', 'Adventure', 'Animation',
                  "Children's", 'Comedy', 'Crime',
                  'Documentary', 'Drama', 'Fantasy',
                  'Film-Noir', 'Horror', 'Musical',
                  'Mystery', 'Romance', 'Sci-Fi',
                  'Thriller', 'War', 'Western']

        # utility functions for importing movies
        def field_to_array(ds, field):
            return hl.cond(ds[field] != 0, hl.array([field]), hl.empty_array(hl.tstr))

        def fields_to_array(ds, fields):
            return hl.flatten(hl.array([field_to_array(ds, f) for f in fields]))

        def rename_columns(ht, new_names):
            return ht.rename({k: v for k, v in zip(ht.row, new_names)})

        info(f'importing users table and writing to {users_path} ...')

        users = rename_columns(
            hl.import_table(user_cluster_readable, key=['f0'], no_header=True, impute=True, delimiter='|'),
            ['id', 'age', 'sex', 'occupation', 'zipcode'])
        users.write(users_path, overwrite=True)

        info(f'importing movies table and writing to {movies_path} ...')

        movies = hl.import_table(movie_cluster_readable, key=['f0'], no_header=True, impute=True, delimiter='|')
        movies = rename_columns(movies,
                                ['id', 'title', 'release date', 'video release date', 'IMDb URL', 'unknown'] + genres)
        movies = movies.drop('release date', 'video release date', 'unknown', 'IMDb URL')
        movies = movies.transmute(genres=fields_to_array(movies, genres))
        movies.write(movies_path, overwrite=True)

        info(f'importing ratings table and writing to {ratings_path} ...')

        ratings = hl.import_table(ratings_cluster_readable, no_header=True, impute=True)
        ratings = rename_columns(ratings,
                                 ['user_id', 'movie_id', 'rating', 'timestamp'])
        ratings = ratings.drop('timestamp')
        ratings.write(ratings_path, overwrite=True)

    else:
        info('Movie Lens files found!')



import numpy as np

import hail as hl
from hail.typecheck import *
from hail.expr.expressions import expr_locus, expr_float64, check_row_indexed
from hail.utils.java import Env


@typecheck(a=np.ndarray,
           radius=oneof(int, float))
def array_windows(a, radius):
    """Returns start and stop indices for window around each array value.

    Examples
    --------

    >>> hl.linalg.utils.array_windows(np.array([1, 2, 4, 4, 6, 8]), 2)
    (array([0, 0, 1, 1, 2, 4]), array([2, 4, 5, 5, 6, 6]))

    >>> hl.linalg.utils.array_windows(np.array([-10.0, -2.5, 0.0, 0.0, 1.2, 2.3, 3.0]), 2.5)
    (array([0, 1, 1, 1, 2, 2, 4]), array([1, 4, 6, 6, 7, 7, 7]))

    Notes
    -----
    For an array ``a`` in ascending order, the resulting ``starts`` and ``stops``
    arrays have the same length as ``a`` and the property that, for all indices
    ``i``, ``[starts[i], stops[i])`` is the maximal range of indices ``j`` such
    that ``a[i] - radius <= a[j] <= a[i] + radius``.

    Index ranges are start-inclusive and stop-exclusive. This function is
    especially useful in conjunction with
    :meth:`.BlockMatrix.sparsify_row_intervals`.

    Parameters
    ----------
    a: :obj:`ndarray` of signed integer or float values
        1-dimensional array of values, non-decreasing with respect to index.
    radius: :obj:`float`
        Non-negative radius of window for values.

    Returns
    -------
    (:class:`ndarray` of :obj:`int64`, :class:`ndarray` of :obj:`int64`)
        Tuple of start indices array and stop indices array.
    """
    if radius < 0:
        raise ValueError(f'array_windows: radius must be non-negative, found {radius}')
    if a.ndim != 1:
        raise ValueError("array_windows: 'a' must be 1-dimensional")
    if not (np.issubdtype(a.dtype, np.signedinteger) or np.issubdtype(a.dtype, np.floating)):
        raise ValueError(f"array_windows: 'a' must be an ndarray of signed integer or float values, "
                         f"found dtype {str(a.dtype)}")

    size = a.size
    if size == 0:
        return np.zeros(shape=0, dtype=np.int64), np.zeros(shape=0, dtype=np.int64)

    if (not np.all(a[:-1] <= a[1:])) or np.isnan(a[0]):
        raise ValueError("array_windows: 'a' must be in ascending order with no nan elements")
    if a[0] - radius > a[0]:
        raise ValueError('array_windows: underflow for a[0] - radius')
    if a[-1] + radius < a[-1]:
        raise ValueError('array_windows: overflow for a[-1] + radius')

    starts, stops = np.zeros(size, dtype=np.int64),  np.zeros(size, dtype=np.int64)
    j, k = 0, 0
    for i in range(size):
        min_val = a[i] - radius
        while j < size and a[j] < min_val:
            j += 1
        starts[i] = j

        max_val = a[i] + radius
        while k < size and a[k] <= max_val:
            k += 1
        stops[i] = k

    return starts, stops


@typecheck(locus_expr=expr_locus(),
           radius=oneof(int, float),
           coord_expr=nullable(expr_float64),
           _localize=bool)
def locus_windows(locus_expr, radius, coord_expr=None, _localize=True):
    """Returns start and stop indices for window around each locus.

    Examples
    --------

    Windows with 2bp radius for one contig with positions 1, 2, 3, 4, 5:

    >>> starts, stops = hl.linalg.utils.locus_windows(
    ...     hl.balding_nichols_model(1, 5, 5).locus,
    ...     radius=2)
    >>> starts, stops
    (array([0, 0, 0, 1, 2]), array([3, 4, 5, 5, 5]))

    The following examples involve three contigs.

    >>> loci = [{'locus': hl.Locus('1', 1), 'cm': 1.0},
    ...         {'locus': hl.Locus('1', 2), 'cm': 3.0},
    ...         {'locus': hl.Locus('1', 4), 'cm': 4.0},
    ...         {'locus': hl.Locus('2', 1), 'cm': 2.0},
    ...         {'locus': hl.Locus('2', 1), 'cm': 2.0},
    ...         {'locus': hl.Locus('3', 3), 'cm': 5.0}]

    >>> ht = hl.Table.parallelize(
    ...         loci,
    ...         hl.tstruct(locus=hl.tlocus('GRCh37'), cm=hl.tfloat64),
    ...         key=['locus'])

    Windows with 1bp radius:

    >>> hl.linalg.utils.locus_windows(ht.locus, 1)
    (array([0, 0, 2, 3, 3, 5]), array([2, 2, 3, 5, 5, 6]))

    Windows with 1cm radius:

    >>> hl.linalg.utils.locus_windows(ht.locus, 1.0, coord_expr=ht.cm)
    (array([0, 1, 1, 3, 3, 5]), array([1, 3, 3, 5, 5, 6]))

    Notes
    -----
    This function returns two 1-dimensional ndarrays of integers,
    ``starts`` and ``stops``, each of size equal to the number of rows.

    By default, for all indices ``i``, ``[starts[i], stops[i])`` is the maximal
    range of row indices ``j`` such that ``contig[i] == contig[j]`` and
    ``position[i] - radius <= position[j] <= position[i] + radius``.

    If the :meth:`.global_position` on `locus_expr` is not in ascending order,
    this method will fail. Ascending order should hold for a matrix table keyed
    by locus or variant (and the associated row table), or for a table that has
    been ordered by `locus_expr`.

    Set `coord_expr` to use a value other than position to define the windows.
    This row-indexed numeric expression must be non-missing, non-``nan``, on the
    same source as `locus_expr`, and ascending with respect to locus
    position for each contig; otherwise the function will fail.

    The last example above uses centimorgan coordinates, so
    ``[starts[i], stops[i])`` is the maximal range of row indices ``j`` such
    that ``contig[i] == contig[j]`` and
    ``cm[i] - radius <= cm[j] <= cm[i] + radius``.

    Index ranges are start-inclusive and stop-exclusive. This function is
    especially useful in conjunction with
    :meth:`.BlockMatrix.sparsify_row_intervals`.

    Parameters
    ----------
    locus_expr : :class:`.LocusExpression`
        Row-indexed locus expression on a table or matrix table.
    radius: :obj:`int`
        Radius of window for row values.
    coord_expr: :class:`.Float64Expression`, optional
        Row-indexed numeric expression for the row value.
        Must be on the same table or matrix table as `locus_expr`.
        By default, the row value is given by the locus position.

    Returns
    -------
    (:class:`ndarray` of :obj:`int64`, :class:`ndarray` of :obj:`int64`)
        Tuple of start indices array and stop indices array.
    """
    if radius < 0:
        raise ValueError(f"locus_windows: 'radius' must be non-negative, found {radius}")
    check_row_indexed('locus_windows', locus_expr)
    if coord_expr is not None:
        check_row_indexed('locus_windows', coord_expr)

    src = locus_expr._indices.source
    if locus_expr not in src._fields_inverse:
        locus = Env.get_uid()
        annotate_fields = {locus: locus_expr}

        if coord_expr is not None:
            if coord_expr not in src._fields_inverse:
                coords = Env.get_uid()
                annotate_fields[coords] = coord_expr
            else:
                coords = src._fields_inverse[coord_expr]

        if isinstance(src, hl.MatrixTable):
            new_src = src.annotate_rows(**annotate_fields)
        else:
            new_src = src.annotate(**annotate_fields)

        locus_expr = new_src[locus]
        if coord_expr is not None:
            coord_expr = new_src[coords]

    if coord_expr is None:
        coord_expr = locus_expr.position

    rg = locus_expr.dtype.reference_genome
    contig_group_expr = hl.agg.group_by(hl.locus(locus_expr.contig, 1, reference_genome=rg), hl.agg.collect(coord_expr))

    # check loci are in sorted order
    last_pos = hl.fold(lambda a, elt: (hl.case()
                                         .when(a <= elt, elt)
                                         .or_error("locus_windows: 'locus_expr' global position must be in ascending order.")),
                       -1,
                       hl.agg.collect(hl.case()
                                        .when(hl.is_defined(locus_expr), locus_expr.global_position())
                                        .or_error("locus_windows: missing value for 'locus_expr'.")))
    checked_contig_groups = (hl.case()
                               .when(last_pos >= 0, contig_group_expr)
                               .or_error("locus_windows: 'locus_expr' has length 0"))

    contig_groups = locus_expr._aggregation_method()(checked_contig_groups, _localize=False)

    coords = hl.sorted(hl.array(contig_groups)).map(lambda t: t[1])
    starts_and_stops = hl._locus_windows_per_contig(coords, radius)

    if not _localize:
        return starts_and_stops

    starts, stops = hl.eval(starts_and_stops)
    return np.array(starts), np.array(stops)


def _check_dims(a, name, ndim, min_size=1):
    if len(a.shape) != ndim:
        raise ValueError(f'{name} must be {ndim}-dimensional, '
                         f'found {a.ndim}')
    for i in range(ndim):
        if a.shape[i] < min_size:
            raise ValueError(f'{name}.shape[{i}] must be at least '
                             f'{min_size}, found {a.shape[i]}')


def _ndarray_matmul_ndim(l, r):
    if l == 1 and r == 1:
        return 0
    elif l == 1:
        return r - 1
    elif r == 1:
        return l - 1
    else:
        assert l == r
        return l



import hail as hl
from hail.typecheck import *


@typecheck(mt=hl.MatrixTable, path=str, batch_size=int, bgzip=bool, header_json_in_file=bool, use_string_key_as_file_name=bool)
def export_entries_by_col(mt: hl.MatrixTable,
                          path: str,
                          batch_size: int = 256,
                          bgzip: bool = True,
                          header_json_in_file: bool = True,
                          use_string_key_as_file_name: bool = False):
    """Export entries of the `mt` by column as separate text files.

    Examples
    --------
    >>> range_mt = hl.utils.range_matrix_table(10, 10)
    >>> range_mt = range_mt.annotate_entries(x = hl.rand_unif(0, 1))
    >>> hl.experimental.export_entries_by_col(range_mt, 'output/cols_files')

    Notes
    -----
    This function writes a directory with one file per column in `mt`. The
    files contain one tab-separated field (with header) for each row field
    and entry field in `mt`. The column fields of `mt` are written as JSON
    in the first line of each file, prefixed with a ``#``.

    The above will produce a directory at ``output/cols_files`` with the
    following files:

    .. code-block:: text

        $ ls -l output/cols_files
        total 80
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 index.tsv
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-00.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-01.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-02.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-03.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-04.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-05.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-06.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-07.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-08.tsv.bgz
        -rw-r--r--  1 hail-dev  wheel  712 Jan 25 17:19 part-09.tsv.bgz

        $ zcat output/cols_files/part-00.tsv.bgz
        #{"col_idx":0}
        row_idx  x
        0        6.2501e-02
        1        7.0083e-01
        2        3.6452e-01
        3        4.4170e-01
        4        7.9177e-02
        5        6.2392e-01
        6        5.9920e-01
        7        9.7540e-01
        8        8.4848e-01
        9        3.7423e-01

    Due to overhead and file system limits related to having large numbers
    of open files, this function will iteratively export groups of columns.
    The `batch_size` parameter can control the size of these groups.

    Parameters
    ----------
    mt : :class:`.MatrixTable`
    path : :obj:`int`
        Path (directory to write to.
    batch_size : :obj:`int`
        Number of columns to write per iteration.
    bgzip : :obj:`bool`
        BGZip output files.
    header_json_in_file : :obj:`bool`
        Include JSON header in each component file (if False, only written to index.tsv)
    """
    if use_string_key_as_file_name and not (len(mt.col_key) == 1 and mt.col_key[0].dtype == hl.tstr):
        raise ValueError(f'parameter "use_string_key_as_file_name" requires a single string column key, found {list(mt.col_key.dtype.values())}')
    hl.utils.java.Env.backend().execute(
        hl.ir.MatrixToValueApply(mt._mir,
                                 {'name': 'MatrixExportEntriesByCol',
                                  'parallelism': batch_size,
                                  'path': path,
                                  'bgzip': bgzip,
                                  'headerJsonInFile': header_json_in_file,
                                  'useStringKeyAsFileName': use_string_key_as_file_name})
    )



import sys

import argparse

from . import connect
from . import describe
from . import diagnose
from . import list_clusters
from . import modify
from . import start
from . import stop
from . import submit


def parser():
    main_parser = argparse.ArgumentParser(
        prog='hailctl dataproc',
        description='Manage and monitor Hail deployments.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    main_parser.add_argument(
        '--beta',
        action='store_true',
        help='Force use of `beta` in gcloud commands')

    subparsers = main_parser.add_subparsers()

    start_parser = subparsers.add_parser(
        'start',
        help='Start a Dataproc cluster configured for Hail.',
        description='Start a Dataproc cluster configured for Hail.')
    submit_parser = subparsers.add_parser(
        'submit',
        help='Submit a Python script to a running Dataproc cluster.',
        description='Submit a Python script to a running Dataproc cluster. To pass arguments to the '
                    'script being submitted, just list them after the name of the script.')
    connect_parser = subparsers.add_parser(
        'connect',
        help='Connect to a running Dataproc cluster.',
        description='Connect to a running Dataproc cluster.')
    diagnose_parser = subparsers.add_parser(
        'diagnose',
        help='Diagnose problems in a Dataproc cluster.',
        description='Diagnose problems in a Dataproc cluster.')
    stop_parser = subparsers.add_parser(
        'stop',
        help='Shut down a Dataproc cluster.',
        description='Shut down a Dataproc cluster.')
    list_parser = subparsers.add_parser(
        'list',
        help='List active Dataproc clusters.',
        description='List active Dataproc clusters.')
    modify_parser = subparsers.add_parser(
        'modify',
        help='Modify active Dataproc clusters.',
        description='Modify active Dataproc clusters.')
    describe_parser = subparsers.add_parser(
        'describe',
        help='Gather information about a hail file (including the schema)',
        description='Gather information about a hail file (including the schema)')

    start_parser.set_defaults(module='start')
    start.init_parser(start_parser)

    submit_parser.set_defaults(module='submit')
    submit.init_parser(submit_parser)

    connect_parser.set_defaults(module='connect')
    connect.init_parser(connect_parser)

    diagnose_parser.set_defaults(module='diagnose')
    diagnose.init_parser(diagnose_parser)

    stop_parser.set_defaults(module='stop')
    stop.init_parser(stop_parser)

    list_parser.set_defaults(module='list')

    modify_parser.set_defaults(module='modify')
    modify.init_parser(modify_parser)

    describe_parser.set_defaults(module='describe')
    describe.init_parser(describe_parser)

    return main_parser


def main(args):
    p = parser()
    if not args:
        p.print_help()
        sys.exit(0)
    jmp = {
        'start': start,
        'submit': submit,
        'connect': connect,
        'diagnose': diagnose,
        'stop': stop,
        'list': list_clusters,
        'modify': modify,
        'describe': describe,
    }

    args, pass_through_args = p.parse_known_args(args=args)
    if "module" not in args:
        p.error('positional argument required')

    jmp[args.module].main(args, pass_through_args)



from .base_ir import *
from .ir import *
from .register_functions import *
from .register_aggregators import *
from .table_ir import *
from .matrix_ir import *
from .blockmatrix_ir import *
from .utils import *
from .matrix_reader import *
from .table_reader import *
from .blockmatrix_reader import *
from .matrix_writer import *
from .table_writer import *
from .blockmatrix_writer import *



from hail.typecheck import typecheck_method
from hail.expr.expressions import unify_types, unify_types_limited, expr_any, \
    expr_bool, ExpressionException, construct_expr, expr_str
from hail import ir


class ConditionalBuilder(object):
    def __init__(self):
        self._ret_type = None
        self._cases = []

    def _unify_type(self, t):
        if self._ret_type is None:
            self._ret_type = t
        else:
            r = unify_types_limited(self._ret_type, t)
            if not r:
                raise TypeError("'then' expressions must have same type, found '{}' and '{}'".format(
                    self._ret_type, t
                ))


class SwitchBuilder(ConditionalBuilder):
    """Class for generating conditional trees based on value of an expression.

    Examples
    --------

    >>> csq = hl.literal('loss of function')
    >>> expr = (hl.switch(csq)
    ...           .when('synonymous', 1)
    ...           .when('SYN', 1)
    ...           .when('missense', 2)
    ...           .when('MIS', 2)
    ...           .when('loss of function', 3)
    ...           .when('LOF', 3)
    ...           .or_missing())
    >>> hl.eval(expr)
    3

    Notes
    -----
    All expressions appearing as the `then` parameters to
    :meth:`~.SwitchBuilder.when` or :meth:`~.SwitchBuilder.default` method
    calls must be the same type.

    See Also
    --------
    :func:`.case`, :func:`.cond`, :func:`.switch`

    Parameters
    ----------
    expr : :class:`.Expression`
        Value to match against.
    """

    @typecheck_method(base=expr_any)
    def __init__(self, base):
        self._base = base
        self._has_missing_branch = False
        super(SwitchBuilder, self).__init__()

    def _finish(self, default):
        assert len(self._cases) > 0

        from hail.expr.functions import cond, bind

        def f(base):
            # build cond chain bottom-up
            expr = default
            for condition, then in self._cases[::-1]:
                expr = cond(condition, then, expr)
            return expr

        return bind(f, self._base)

    @typecheck_method(value=expr_any, then=expr_any)
    def when(self, value, then) -> 'SwitchBuilder':
        """Add a value test. If the `base` expression is equal to `value`, then
         returns `then`.

        Warning
        -------
        Missingness always compares to missing. Both ``NA == NA`` and
        ``NA != NA`` return ``NA``. Use :meth:`~SwitchBuilder.when_missing`
        to test missingness.

        Parameters
        ----------
        value : :class:`.Expression`
        then : :class:`.Expression`

        Returns
        -------
        :class:`.SwitchBuilder`
            Mutates and returns `self`.
        """
        can_compare = unify_types(self._base.dtype, value.dtype)
        if not can_compare:
            raise TypeError("cannot compare expressions of type '{}' and '{}'".format(
                self._base.dtype, value.dtype))

        self._unify_type(then.dtype)
        self._cases.append((self._base == value, then))
        return self

    @typecheck_method(then=expr_any)
    def when_missing(self, then) -> 'SwitchBuilder':
        """Add a test for missingness. If the `base` expression is missing,
        returns `then`.

        Parameters
        ----------
        then : :class:`.Expression`

        Returns
        -------
        :class:`.SwitchBuilder`
            Mutates and returns `self`.
        """
        if self._has_missing_branch:
            raise ExpressionException("'when_missing' can only be called once")
        self._unify_type(then.dtype)

        from hail.expr.functions import is_missing
        # need to insert at 0, because upstream missingness would propagate
        self._cases.insert(0, (is_missing(self._base), then))
        return self

    @typecheck_method(then=expr_any)
    def default(self, then):
        """Finish the switch statement by adding a default case.

        Notes
        -----
        If no value from a :meth:`~.SwitchBuilder.when` call is matched, then
        `then` is returned.

        Parameters
        ----------
        then : :class:`.Expression`

        Returns
        -------
        :class:`.Expression`
        """
        if len(self._cases) == 0:
            return then
        self._unify_type(then.dtype)
        return self._finish(then)

    def or_missing(self):
        """Finish the switch statement by returning missing.

        Notes
        -----
        If no value from a :meth:`~.SwitchBuilder.when` call is matched, then
        the result is missing.

        Parameters
        ----------
        then : :class:`.Expression`

        Returns
        -------
        :class:`.Expression`
        """
        if len(self._cases) == 0:
            raise ExpressionException("'or_missing' cannot be called without at least one 'when' call")
        from hail.expr.functions import null
        return self._finish(null(self._ret_type))


class CaseBuilder(ConditionalBuilder):
    """Class for chaining multiple if-else statements.


    Examples
    --------

    >>> x = hl.literal('foo bar baz')
    >>> expr = (hl.case()
    ...           .when(x[:3] == 'FOO', 1)
    ...           .when(x.length() == 11, 2)
    ...           .when(x == 'secret phrase', 3)
    ...           .default(0))
    >>> hl.eval(expr)
    2

    Notes
    -----
    All expressions appearing as the `then` parameters to
    :meth:`~.CaseBuilder.when` or :meth:`~.CaseBuilder.default` method calls
    must be the same type.

    Parameters
    ----------
    missing_false: :obj:`.bool`
        Treat missing predicates as ``False``.

    See Also
    --------
    :func:`.case`, :func:`.cond`, :func:`.switch`
    """

    def __init__(self, missing_false=False):
        super(CaseBuilder, self).__init__()
        self._missing_false = missing_false

    def _finish(self, default):
        assert len(self._cases) > 0

        from hail.expr.functions import cond

        expr = default
        for conditional, then in self._cases[::-1]:
            expr = cond(conditional, then, expr, missing_false=self._missing_false)
        return expr

    @typecheck_method(condition=expr_bool, then=expr_any)
    def when(self, condition, then) -> 'CaseBuilder':
        """Add a branch. If `condition` is ``True``, then returns `then`.

        Warning
        -------
        Missingness is treated similarly to :func:`.cond`. Missingness is
        **not** treated as ``False``. A `condition` that evaluates to missing
        will return a missing result, not proceed to the next case. Always
        test missingness first in a :class:`.CaseBuilder`.

        Parameters
        ----------
        condition: :class:`.BooleanExpression`
        then : :class:`.Expression`

        Returns
        -------
        :class:`.CaseBuilder`
            Mutates and returns `self`.
        """
        self._unify_type(then.dtype)
        self._cases.append((condition, then))
        return self

    @typecheck_method(then=expr_any)
    def default(self, then):
        """Finish the case statement by adding a default case.

        Notes
        -----
        If no condition from a :meth:`~.CaseBuilder.when` call is ``True``,
        then `then` is returned.

        Parameters
        ----------
        then : :class:`.Expression`

        Returns
        -------
        :class:`.Expression`
        """
        if len(self._cases) == 0:
            return then
        self._unify_type(then.dtype)
        return self._finish(then)

    def or_missing(self):
        """Finish the case statement by returning missing.

        Notes
        -----
        If no condition from a :meth:`.CaseBuilder.when` call is ``True``, then
        the result is missing.

        Parameters
        ----------
        then : :class:`.Expression`

        Returns
        -------
        :class:`.Expression`
        """
        if len(self._cases) == 0:
            raise ExpressionException("'or_missing' cannot be called without at least one 'when' call")
        from hail.expr.functions import null
        return self._finish(null(self._ret_type))

    @typecheck_method(message=expr_str)
    def or_error(self, message):
        """Finish the case statement by throwing an error with the given message.

        Notes
        -----
        If no condition from a :meth:`.CaseBuilder.when` call is ``True``, then
        an error is thrown.

        Parameters
        ----------
        message : :class:`.Expression` of type :data:`tstr`

        Returns
        -------
        :class:`.Expression`
        """
        if len(self._cases) == 0:
            raise ExpressionException("'or_error' cannot be called without at least one 'when' call")
        error_expr = construct_expr(ir.Die(message._ir, self._ret_type), self._ret_type)
        return self._finish(error_expr)



import hail as hl


def full_outer_join_mt(left: hl.MatrixTable, right: hl.MatrixTable) -> hl.MatrixTable:
    """Performs a full outer join on `left` and `right`.

    Replaces row, column, and entry fields with the following:

     - `left_row` / `right_row`: structs of row fields from left and right.
     - `left_col` / `right_col`: structs of column fields from left and right.
     - `left_entry` / `right_entry`: structs of entry fields from left and right.

    Parameters
    ----------
    left : :class:`.MatrixTable`
    right : :class:`.MatrixTable`

    Returns
    -------
    :class:`.MatrixTable`
    """

    if [x.dtype for x in left.row_key.values()] != [x.dtype for x in right.row_key.values()]:
        raise ValueError(f"row key types do not match:\n"
                         f"  left:  {list(left.row_key.values())}\n"
                         f"  right: {list(right.row_key.values())}")

    if [x.dtype for x in left.col_key.values()] != [x.dtype for x in right.col_key.values()]: 
        raise ValueError(f"column key types do not match:\n"
                         f"  left:  {list(left.col_key.values())}\n"
                         f"  right: {list(right.col_key.values())}")

    left = left.select_rows(left_row=left.row)
    left_t = left.localize_entries('left_entries', 'left_cols')
    right = right.select_rows(right_row=right.row)
    right_t = right.localize_entries('right_entries', 'right_cols')

    ht = left_t.join(right_t, how='outer')
    ht = ht.annotate_globals(
        left_keys=hl.group_by(
            lambda t: t[0],
            hl.zip_with_index(
                ht.left_cols.map(lambda x: hl.tuple([x[f] for f in left.col_key])), index_first=False)).map_values(
            lambda elts: elts.map(lambda t: t[1])),
        right_keys=hl.group_by(
            lambda t: t[0],
            hl.zip_with_index(
                ht.right_cols.map(lambda x: hl.tuple([x[f] for f in right.col_key])), index_first=False)).map_values(
            lambda elts: elts.map(lambda t: t[1])))
    ht = ht.annotate_globals(
        key_indices=hl.array(ht.left_keys.key_set().union(ht.right_keys.key_set()))
            .map(lambda k: hl.struct(k=k, left_indices=ht.left_keys.get(k), right_indices=ht.right_keys.get(k)))
            .flatmap(lambda s: hl.case()
                     .when(hl.is_defined(s.left_indices) & hl.is_defined(s.right_indices),
                           hl.range(0, s.left_indices.length()).flatmap(
                               lambda i: hl.range(0, s.right_indices.length()).map(
                                   lambda j: hl.struct(k=s.k, left_index=s.left_indices[i],
                                                       right_index=s.right_indices[j]))))
                     .when(hl.is_defined(s.left_indices),
                           s.left_indices.map(
                               lambda elt: hl.struct(k=s.k, left_index=elt, right_index=hl.null('int32'))))
                     .when(hl.is_defined(s.right_indices),
                           s.right_indices.map(
                               lambda elt: hl.struct(k=s.k, left_index=hl.null('int32'), right_index=elt)))
                     .or_error('assertion error')))
    ht = ht.annotate(__entries=ht.key_indices.map(lambda s: hl.struct(left_entry=ht.left_entries[s.left_index],
                                                                      right_entry=ht.right_entries[s.right_index])))
    ht = ht.annotate_globals(__cols=ht.key_indices.map(
        lambda s: hl.struct(**{f: s.k[i] for i, f in enumerate(left.col_key)},
                            left_col=ht.left_cols[s.left_index],
                            right_col=ht.right_cols[s.right_index])))
    ht = ht.drop('left_entries', 'left_cols', 'left_keys', 'right_entries', 'right_cols', 'right_keys', 'key_indices')
    return ht._unlocalize_entries('__entries', '__cols', list(left.col_key))
